<?php

namespace apexl\Io\modules\invoice\entities;

use apexl\Io\includes\Entity;
use apexl\Io\includes\System;
use apexl\Io\includes\Utils;
use apexl\Io\modules\invoice\entities\operators\invoicePaymentOperator;
use apexl\Io\modules\invoice\entities\operators\invoiceSubscriptionOperator;

class invoiceSubscriptionEntity extends Entity{

    public function __construct()
    {
        parent::__construct('invoice_subscription');
        $this->setOperator(new invoiceSubscriptionOperator('invoice_subscription'));
    }

}